// ************************************************************
//					H i F i - P R E A M P
// ************************************************************
//
// 	Autor:			Mario Richter
// 	Dateiname:	preamp.c
// 	Projekt:		HiFi Audio Preamplifier
// 	Datum:			Mrz 2009
// 	Target-MCU:	ATmega8
//
// BESCHREIBUNG
//
// 	Software fr den HiFi Audio Preamplifier. Die Software kmmert
// 	sich um die Ansteuerung des Displays, wertet die Eingangssignale
//  vom IR-Sensor und von den Tastern aus und steuert die Relais der
//  Eingangsmatrix sowie den Treiber fr das Motorpoti.
//
// PORTBELEGUNG
//
//  Relaismatrix:   PD0..PD6    Kanal1..7     (H = Kanal ausgewhlt)
//
//  Display:        PB2         D4            Daten (4 Bit-Mode)
//                  PB1         D5
//                  PB0         D6
//                  PD7         D7
//                  PB3         E             Enable
//                  PB4         RW            Read/Write
//                  PB5         RS            Register Select
//
//  Taster:         PC1         Mute          Stummschaltung
//                  PC2         Source        Kanalauswahl
//
//  Motorsteuerung: PC3         MOT_ENABLE    Enable
//                  PC4         MOT_LEFT      Linkslauf (leiser)
//                  PC5         MOT_RIGHT     Rechtslauf (lauter)
//
//  IR-Sensor:      PC0         IR            Ausgang TSOP1736
//
//  Mute-Schaltung: PB7         MUTE          H = Mute
//
// ************************************************************

#include <avr/io.h>					          // I/O Ports
#include <avr/interrupt.h>            // Interrupts
#include <stdint.h>					          // Standard-Datentypen
#include <avr/eeprom.h>

#include "dogm.h"                     // Routinen fr DOGM LCD
#include "rc5.h"                      // Routinen fr RC5 IR Receiver

#define DDR(x) (*(&x - 1))      /* address of data direction register of port x */
#if defined(__AVR_ATmega64__) || defined(__AVR_ATmega128__)
    /* on ATmega64/128 PINF is on port 0x00 and not 0x60 */
    #define PIN(x) ( &PORTF==&(x) ? _SFR_IO8(0x00) : (*(&x - 2)) )
#else
  #define PIN(x) (*(&x - 2))    /* address of input register of port x          */
#endif

// delay fr Ticker
#define TICK_10MS     (uint8_t)(F_CPU * 10e-3 / 512)
#define TICK_1MS      (uint8_t)(F_CPU * 1e-3 / 512)

#define MENU_DELAY    200     // 200 * 10ms = 2s

// Port defines
#define SOURCE_PORT   PORTD

#define MUTE_PORT     PORTB
#define MUTE          PB7

#define SWITCH_PORT   PORTC
#define SWITCH_MUTE   PC1
#define SWITCH_SRC    PC2

#define MOTOR_PORT    PORTC
#define MOTOR_ENA     PC3
#define MOTOR_LEFT    PC4
#define MOTOR_RIGHT   PC5

// Source-Anzahl
#define MAX_SOURCE        5

// Tape Makros
#define TAPE_ON           { SOURCE_PORT |=  (1<<PD6); }
#define TAPE_OFF          { SOURCE_PORT &= ~(1<<PD6); }

// Entprellzhler
#define SWITCH_DEBOUNCE_CNTR    10    // 10 * 1ms = 10ms
#define SWITCH_LONG_PUSH        1000  // 1000 * 1ms = 1s

// RC5 Codes
#define RC5_VOLUME_UP   16            // Lautstrke +
#define RC5_VOLUME_DN   17            // Lautstrke -
#define RC5_CHANNEL_UP  32            // Kanal +
#define RC5_CHANNEL_DN  33            // Kanal -
#define RC5_MUTE        13            // Mute

#define RC5_DELAY       20            // RC5 Kommandos mssen 20*10ms = 200ms auseinanderliegen
#define EEPROM_DELAY    1000          // 1000 * 10ms = 10s
#define CHANGE_DELAY    2             // 2*10ms = 20ms (zweimal verwendet fr Ein- + Ausblenden)

uint8_t ticker10ms;                   // 10ms ticker
uint8_t ticker1ms;                    // 1ms ticker
uint16_t menu_delay;                  // Menverzgerung

uint8_t menu;                         // dargestelltes Men (0=main, 1=Quellenwahl)
uint8_t source, source_new;           // Quellenauswahl (0..5, 0=CD, 1=Tuner...)
uint8_t muting, muting_new;           // Muting an (1) oder aus (0)
uint8_t tape, tape_new;               // Tape playback (1) oder Bypass (0)

uint8_t tape_eep;                     // EEPROM Variablen
uint8_t source_eep;
uint8_t mute_eep;

uint16_t tape_eep_dly;
uint16_t source_eep_dly;
uint16_t mute_eep_dly;

uint8_t menu_dsp, source_dsp;         // Displayparameter: zeigt welche Parameter aktuell
uint8_t muting_dsp, tape_dsp;         // auf dem LCD dargestellt werden
uint8_t vol_stat_dsp;                 // =1: Refresh ntig

uint8_t mute_dbcnt;                   // Entprellzhler fr Mute-Taster
uint16_t src_dbcnt;                   // Entprellzhler fr Source-Taster (+ Erkennung langer Tastendruck)

uint8_t vol_up_delay;                 // 100ms delay counter fr Lautstrke +
uint8_t vol_dn_delay;                 // 100ms delay counter fr Lautstrke -

uint8_t source_delay;                 // delay fr Quellenwechsel
uint8_t muting_delay;                 // delay fr Muting
uint8_t tape_delay;                   // delay fr Tape

uint8_t rc5_delay;                    // delay counter fr RC5-Befehle


// Hauptroutine
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

int main(void)
{
  unsigned int i;

  vol_up_delay = 0;
  vol_dn_delay = 0;
  rc5_delay = 0;

  source_eep_dly = 0;
  tape_eep_dly = 0;
  mute_eep_dly = 0;

  menu_dsp = 99;                      // erzwingt Display-Refresh

  // Ports initialisieren
  DDR(SOURCE_PORT)  = 0x7F;           // PD6..PD0 als Ausgnge (Relais)
  SOURCE_PORT       = 0x00;           // Alle Relais aus

  DDR(MOTOR_PORT)   = 0x38;           // PC3..PC5 als Ausgnge (Motor)
  MOTOR_PORT        = 0x00;           // Alle aus

  DDR(MUTE_PORT)    = 0x80;           // PB7 als Ausgnge (Mute)
  MUTE_PORT         = 0x80;           // Mute anfangs an (!)        

  // Timer0 fr RC5-Receiver vorbereiten
  TCCR0 = 1<<CS02;                    // teile durch 256
  TIMSK = 1<<TOIE0;                   // enable interrupt

  // Display initialisieren und Startmeldung zeigen
  dogm_init();
  dogm_puts("HiFi Preamp V1.0");
  dogm_gotoxy(0, 1);
  dogm_puts("  mario001.de   ");

  // warte 3s
  _delay_ms(3000.0);
  dogm_clear();

  // letzte Einstellungen aus dem EEPROM holen
  source_eep = eeprom_read_byte(0x00);
  if(source_eep >= MAX_SOURCE)
  {
    // Keine Daten im EEPROM vorhanden, wir nehmen Standardwerte
    source = 0;
    source_new = 0;

    SOURCE_PORT &= ~(0x7F);
    SOURCE_PORT |= 0x01;
  }
  else
  {
    source = source_eep;
    source_new = source;

    switch(source)            // Kanal entsprechend Vorgabe auswhlen
    {
    case 0:   SOURCE_PORT &= ~(0x3E);   SOURCE_PORT |= 0x01;  break;
    case 1:   SOURCE_PORT &= ~(0x3D);   SOURCE_PORT |= 0x02;  break;
    case 2:   SOURCE_PORT &= ~(0x3B);   SOURCE_PORT |= 0x04;  break;
    case 3:   SOURCE_PORT &= ~(0x37);   SOURCE_PORT |= 0x08;  break;
    case 4:   SOURCE_PORT &= ~(0x2F);   SOURCE_PORT |= 0x10;  break;
    case 5:   SOURCE_PORT &= ~(0x1F);   SOURCE_PORT |= 0x20;  break;
    default:  SOURCE_PORT &= ~(0x3F);
    }
  }  

  mute_eep = eeprom_read_byte(0x01);
  if(mute_eep == 1)
  {
    muting = 1;
    muting_new = 1;
  }
  else
  {
    muting = 0;
    muting_new = 0;
  }

  tape_eep = eeprom_read_byte(0x02);
  if(tape_eep == 1)
  {
    tape = 1;
    tape_new = 1;
  }
  else
  {
    tape = 0;
    tape_new = 0;
  }

  // nach nochmal 10ms Mute aufheben
  _delay_ms(10.0);
  MUTE_PORT &= ~(1<<MUTE);

  sei();

  while(1)
  {
    cli();
    i = rc5.data & 0x3f;     // read two bytes from interrupt !
    rc5.data = 0;
    sei();


    // RC5-Kommandos decodieren
    // -----------------------------
    switch(i)
    {
    case RC5_VOLUME_UP:         // LAUTSTRKE +
      vol_up_delay = 15;        // Kommando halten fr 15*10ms = 150ms
      vol_dn_delay = 0;
      vol_stat_dsp = 1;         // Display bentigt Refresh
      break;

    case RC5_VOLUME_DN:         // LAUTSTRKE -
      vol_dn_delay = 15;        // Kommando halten fr 15*10ms = 150ms
      vol_up_delay = 0;
      vol_stat_dsp = 1;         // Display bentigt Refresh
      break;

    case RC5_CHANNEL_UP:            // KANAL +
      if(!rc5_delay)                // nur ausfhren wenn letztes Kommando mehr als 200ms vorher
      {
        rc5_delay = RC5_DELAY;      // delay counter setzen
        menu = 1;                   // Untermen im Display anwhlen
        menu_delay = MENU_DELAY;    // Men-Delay setzen

        source_eep_dly = 0;
                
        if(source_new == MAX_SOURCE)  // Quelle weiterschalten, bei berlauf zurck auf 0
          source_new = 0;
        else
          source_new += 1;
      }
      break;

    case RC5_CHANNEL_DN:            // KANAL -
      if(!rc5_delay)                // nur ausfhren wenn letztes Kommando mehr als 200ms vorher
      {
        rc5_delay = RC5_DELAY;      // delay counter setzen
        menu = 1;                   // Untermen im Display anwhlen
        menu_delay = MENU_DELAY;    // Men-Delay setzen

        source_eep_dly = 0;
        
        if(source_new == 0)         // Quelle weiterschalten, bei Unterlauf zurck auf Maximum
          source_new = MAX_SOURCE;
        else
          source_new -= 1;
      }
      break;

    case RC5_MUTE:                  // MUTE
      if(!rc5_delay)                // nur ausfhren wenn letztes Kommando mehr als 200ms vorher
      {
        rc5_delay = RC5_DELAY;      // delay counter setzen
        muting_new ^= 0x01;         // Muting toggeln
      }
    }

    // Motoransteuerung
    // -----------------------------
    if(vol_dn_delay)            // vol_dn_delay -> Motor links drehen
    {
      MOTOR_PORT &= ~(1<<MOTOR_RIGHT);
      MOTOR_PORT |= ((1<<MOTOR_ENA) | (1<<MOTOR_LEFT));
    }
    else if(vol_up_delay)       // vol_up_delay -> Motor rechts drehen
    {
      MOTOR_PORT &= ~(1<<MOTOR_LEFT);
      MOTOR_PORT |= ((1<<MOTOR_ENA) | (1<<MOTOR_RIGHT));
    }
    else                        // Motor aus
      MOTOR_PORT &= ~((1<<MOTOR_ENA) | (1<<MOTOR_LEFT) | (1<<MOTOR_RIGHT));

    // Quellenwahl
    // -----------------------------
    if(muting)
      SOURCE_PORT &= ~(0x7F);   // Muting aktiv, dann alle Relais aus
    else
    {
      switch(source)            // sonst Relais entsprechend der gewhlten Quelle schalten
      {
      case 0:   SOURCE_PORT &= ~(0x3E);   SOURCE_PORT |= 0x01;  break;
      case 1:   SOURCE_PORT &= ~(0x3D);   SOURCE_PORT |= 0x02;  break;
      case 2:   SOURCE_PORT &= ~(0x3B);   SOURCE_PORT |= 0x04;  break;
      case 3:   SOURCE_PORT &= ~(0x37);   SOURCE_PORT |= 0x08;  break;
      case 4:   SOURCE_PORT &= ~(0x2F);   SOURCE_PORT |= 0x10;  break;
      case 5:   SOURCE_PORT &= ~(0x1F);   SOURCE_PORT |= 0x20;  break;
      default:  SOURCE_PORT &= ~(0x3F);
      }

      if(tape)                  // falls Tape aktiv, Tape-Relais schalten
        SOURCE_PORT |= 0x40;
      else
        SOURCE_PORT &= ~(0x40);
    }

    // 1ms-Timer
    // -----------------------------
    if(ticker1ms >= TICK_1MS)
    {
      ticker1ms = 0;

      // Mute-Taster entprellen
      if(!(PIN(SWITCH_PORT) & (1<<SWITCH_MUTE)))
      {
        if(mute_dbcnt <= SWITCH_DEBOUNCE_CNTR)
          mute_dbcnt++;

        if(mute_dbcnt == SWITCH_DEBOUNCE_CNTR)
          muting_new ^= 0x01;
      }
      else
        mute_dbcnt = 0;

      // Source-Taster entprellen
      if(!(PIN(SWITCH_PORT) & (1<<SWITCH_SRC)))
      {
        if(src_dbcnt <= SWITCH_LONG_PUSH)
          src_dbcnt++;

        if(src_dbcnt == SWITCH_LONG_PUSH)   // Langer Tastendruck toggelt Tape-Betrieb
        {
          tape_new ^= 0x01;
          tape_eep_dly = 0;
        }
      }
      else
      {
        // es gab einen KURZEN Tastendruck
        if((src_dbcnt > SWITCH_DEBOUNCE_CNTR) && (src_dbcnt < SWITCH_LONG_PUSH))
        {
          menu = 1;                     // Untermen fr Quellenauswahl aktivieren
          menu_delay = MENU_DELAY;      // Mendelay setzen

          source_eep_dly = 0;
          
          if(source_new == MAX_SOURCE)  // Quelle weiterschalten, bei berlauf zurck auf 0
            source_new = 0;
          else
            source_new += 1;
        }

        src_dbcnt = 0;                  // Entprellzhler resetten
      }
    }

    // 10ms-Timer
    // -----------------------------
    if(ticker10ms >= TICK_10MS)
    {
      ticker10ms = 0;

      // Bevor Quelle gewechselt wird, Mute-Schaltung aktivieren
      if((source != source_new) && (source_delay == 0))
      {
        source_delay = CHANGE_DELAY;
        MUTE_PORT |= (1<<MUTE);
      }

      // Quellenumschaltung
      if(source_delay)
      {
        source_delay--;

        if(source_delay == 0)
        {
          // SCHRITT 1: Auf neue Quelle umschalten
          if(source != source_new)
          {
            source = source_new;
            source_delay = CHANGE_DELAY;
          }
          // SCHRITT 2: Mute-Schaltung deaktivieren
          else
          {
            MUTE_PORT &= ~(1<<MUTE);
          }
        }
      }      

      // Muting umschalten
      if((muting != muting_new) && (muting_delay == 0))
      {
        muting_delay = CHANGE_DELAY;
        MUTE_PORT |= (1<<MUTE);

        // Soll Mute deaktiviert werden, dann Relais gleich aus 
        // (Mute-Schaltung ist sowieso bereits aktiv)
        if(muting_new == 0)
          muting = 0;
      }

      // Muting
      if(muting_delay)
      {
        if(--muting_delay == 0)
        {
          muting = muting_new;

          // Wenn Muting jetzt aus, dann Mute-Schaltung deaktivieren
          if(muting == 0)
            MUTE_PORT &= ~(1<<MUTE);
        }
      }

      // Tape-Umschaltung vorbereiten (Mute-Schaltung aktiv)
      if((tape != tape_new) && (tape_delay == 0))
      {
        tape_delay = 4;
        MUTE_PORT |= (1<<MUTE);
      }

      // Tape-Umschaltung
      if(tape_delay)
      {
        if(--tape_delay == 0)
        {
          // SCHRITT 1: Tape umschalten
          if(tape != tape_new)
          {
            tape = tape_new;
            tape_delay = 4;
          }
          // SCHRITT 2: Mute-Schaltung deaktivieren
          else
          {
            MUTE_PORT &= ~(1<<MUTE);
          }
        }
      }
      
      // Timer fr Lautstrke + Delay fhren
      if(vol_up_delay > 0)
      {
        vol_up_delay--;

        if(vol_up_delay == 0)           // Wenn jetzt 0, dann Display updaten
          vol_stat_dsp = 1;
      }

      // Timer fr Lautstrke - Delay fhren
      if(vol_dn_delay > 0)
      {
        vol_dn_delay--;

        if(vol_dn_delay == 0)           // Wenn jetzt 0, dann Display updaten
          vol_stat_dsp = 1;
      }

      // Untermen wird angezeigt
      if(menu_delay)
      {
        menu_delay--;

        if(menu_delay == 0)             // Wenn Delay vorbei, zurck ins Hauptmen
        {
          //source = source_new;
          menu = 0;
        }
      }

      // RC5-Befehlsverzgerung
      if(rc5_delay)
        rc5_delay--;

      // EEPROM-Schreibvorgnge
      if(source != source_eep)          // Neue Quelle eingestellt ?
      {
        if(++source_eep_dly == EEPROM_DELAY)    // Dann nach dem Delay ins EEPROM eintragen
        {
          eeprom_write_byte(0x00, source);
          source_eep = source;
          source_eep_dly = 0;
        }
      }

      if(tape != tape_eep)              // Neuer Tape-Zustand eingestellt ?
      {
        if(++tape_eep_dly == EEPROM_DELAY)      // Dann nach dem Delay ins EEPROM eintragen
        {
          eeprom_write_byte(0x02, tape);
          tape_eep = tape;
          tape_eep_dly = 0;
        }
      }
    }

    // Display-Handler: Hauptmen
    // -----------------------------
    if(menu == 0)
    {
      if(menu != menu_dsp)      // Wird gerade nicht das Hauptmen angezeigt, alles neu zeichnen
      {
        dogm_clear();

        switch(source)          // Ausgewhlte Quelle darstellen
        {
        case 0:   dogm_puts("       CD       ");  break;
        case 1:   dogm_puts("      Tuner     ");  break;
        case 2:   dogm_puts("      Aux1      ");  break;
        case 3:   dogm_puts("      Aux2      ");  break;
        case 4:   dogm_puts("       PC       ");  break;
        case 5:   dogm_puts("      IPod      ");  break;
        }

        dogm_gotoxy(0, 1);

        if(tape)                // Tape-Status darstellen
          dogm_puts("TAPE  ");
        else
          dogm_puts("      ");

        if(vol_dn_delay)        // Lautstrkenderungen ber Fernbedienung anzeigen
          dogm_puts("VOL-  ");
        else if(vol_up_delay)
          dogm_puts("VOL+  ");
        else
          dogm_puts("      ");

        if(muting)              // Mute-Status anzeigen
          dogm_puts("MUTE");

        menu_dsp = menu;        // Merken, was gerade angezeigt wird
        source_dsp = source;
        tape_dsp = tape;
        muting_dsp = muting;
      }
      else                      // Hauptmen wird bereits angezeigt, prfen was aktualisiert werden muss
      {
        if(source_dsp != source)  // Prfe ob Quelle passt; wenn nicht, dann neu zeichnen
        {
          source_dsp = source;  // Als upgedatet markieren
          dogm_gotoxy(0, 0);

          switch(source)        // Ausgewhlte Quelle darstellen
          {
          case 0:   dogm_puts("       CD       ");  break;
          case 1:   dogm_puts("      Tuner     ");  break;
          case 2:   dogm_puts("      Aux1      ");  break;
          case 3:   dogm_puts("      Aux2      ");  break;
          case 4:   dogm_puts("       PC       ");  break;
          case 5:   dogm_puts("      IPod      ");  break;
          }
        }

        if(tape_dsp != tape)    // Prfe ob Tape-Status passt
        {
          tape_dsp = tape;
          dogm_gotoxy(0, 1);

          if(tape)              // Tape-Status neuzeichnen
            dogm_puts("TAPE  ");
          else
            dogm_puts("      ");
        }

        if(vol_stat_dsp)        // Flag zeigt, ob Vol+/Vol- neu gezeichnet werden muss
        {
          vol_stat_dsp = 0;
          dogm_gotoxy(6, 1);

          if(vol_dn_delay)      // VOL-Kommandos von der Fernbedienung darstellen
            dogm_puts("VOL-  ");
          else if(vol_up_delay)
            dogm_puts("VOL+  ");
          else
            dogm_puts("      ");
        }

        if(muting_dsp != muting)  // Prfe ob Muting passt
        {
          muting_dsp = muting;
          dogm_gotoxy(12, 1);

          if(muting)              // Muting zeigen
            dogm_puts("MUTE");
          else
            dogm_puts("    ");
        }
      }
    }

    // Displayhandler: Quellenmen
    // -----------------------------
    if(menu == 1)
    {
      if(menu_dsp != menu)        // Quellenauswahlmen wird noch nicht angezeigt
      {
        source_dsp = source_new;  // Neue zu whlende Quelle darstellen
        menu_dsp = menu;

        dogm_clear();
        dogm_puts(" Select Source: ");

        dogm_gotoxy(0, 1);
        switch(source_new)        // Ausgewhlte Quelle darstellen
        {
        case 0:   dogm_puts("       CD       ");  break;
        case 1:   dogm_puts("      Tuner     ");  break;
        case 2:   dogm_puts("      Aux1      ");  break;
        case 3:   dogm_puts("      Aux2      ");  break;
        case 4:   dogm_puts("       PC       ");  break;
        case 5:   dogm_puts("      IPod      ");  break;
        }
      }
      else                        // Quellenauswahlmen wird schon angezeigt
      {
        if(source_dsp != source_new)
        {
          source_dsp = source_new;
          dogm_gotoxy(0, 1);

          switch(source_new)      // Auszuwhlende Quelle anzeigen
          {
          case 0:   dogm_puts("       CD       ");  break;
          case 1:   dogm_puts("      Tuner     ");  break;
          case 2:   dogm_puts("      Aux1      ");  break;
          case 3:   dogm_puts("      Aux2      ");  break;
          case 4:   dogm_puts("       PC       ");  break;
          case 5:   dogm_puts("      IPod      ");  break;
          }
        }
      }
    }
  }
}

// Timer0 ISR
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ISR(SIG_OVERFLOW0)
{
  ticker1ms++;
  ticker10ms++;

  rc5_scheduler();
}

